from dataclasses import dataclass
@dataclass
class AccessToken:
    consumerID: str
    consumerSecret: str
    twoFactorType: int
    code: str
    isSave: bool = True
@dataclass
class GetOTP:
    consumerID: str
    consumerSecret: str
@dataclass
class NewOrder:
    account: str
    requestID: str
    instrumentID: str
    market: str
    buySell: str
    orderType: str
    price: float
    quantity: int
    stopOrder: bool = False
    stopPrice: float = 0.0
    stopType: str = ''
    stopStep: float = 0.0
    lossStep: float = 0.0
    profitStep: float = 0.0
    channelID: str = 'TA'
    code: str = ''
    deviceId: str = ''
    userAgent: str = ''

@dataclass
class CancelOrder:
    account: str
    requestID: str
    orderID: str
    marketID: str
    instrumentID: str
    buySell: str
    code: str = ''
    deviceId: str = ''
    userAgent: str = ''

@dataclass
class ModifyOrder:
    account: str
    requestID: str
    orderID: str
    marketID: str
    instrumentID: str
    price: float
    quantity: int
    buySell: str
    orderType: str
    code: str = ''
    deviceId: str = ''
    userAgent: str = ''

@dataclass
class StockAccountBalance:
    account: str

@dataclass
class DerivativeAccountBalance:
    account: str

@dataclass
class PPMMRAccount:
    account: str

@dataclass
class StockPosition:
    account: str

@dataclass
class DerivativePosition:
    account: str
    querySummary: bool = True

@dataclass
class MaxBuyQty:
    account: str
    instrumentID: str
    price: float

@dataclass
class MaxSellQty:
    account: str
    instrumentID : str
    price: str

@dataclass
class OrderHistory:
    account: str
    startDate: str
    endDate: str
    
@dataclass
class OrderBook:
    account: str



from .constant import api
import time
# import gevent.monkey
# gevent.monkey.patch_all()
from requests import Session
from signalr import Connection
from . import FCTradingClient


class FCTradingStream(object):

	def __init__(self, fctrading_client: FCTradingClient, stream_url: str, on_message, on_error, last_notify_id: str = "-1"):
		self.client = fctrading_client
		self._stream_url = stream_url
		self._fc_handlers = []
		self._fc_handlers.append(on_message)
		self._fc_error_handlers = []
		self._fc_error_handlers.append(on_error)
		self.lastId = last_notify_id

	def fc_on_message(self, fc_message):
		for fc_handler in self._fc_handlers:
			fc_handler(fc_message)

	def fc_on_error(self, fc_error):

		for fc_error_handler in self._fc_error_handlers:
			fc_error_handler(fc_error)

	def start(self):
		with Session() as fc_session:

			fc_session.headers["Authorization"] = "Bearer " + self.client.get_access_token()
			fc_session.headers["NotifyID"] = self.lastId			
			connection = Connection(self._stream_url + api.SIGNALR, fc_session)
			
			chat = connection.register_hub(api.SIGNALR_HUB_FC)
			chat.client.on(api.SIGNALR_METHOD, self.fc_on_message)
			chat.client.on(api.SIGNALR_METHOD_ERROR, self.fc_on_error)
			connection.error += self.fc_on_error
			
			with connection:
				while True:
					time.sleep(10)





import random
import string


class core_helper(object):
	def __init__(self):
		self.ping = 'pong'

	@staticmethod
	def random_string(length):
		stringy = string.ascii_lowercase
		return ''.join(random.choice(stringy) for i in range(length))

	@staticmethod
	def to_lower(json_object):
		new = {}
		for k, v in json_object.items():
			if isinstance(v, dict):
				v = core_helper.to_lower(v)
			new[k.lower()] = v
		return new




SIGNALR = "v2.0/signalr"
SIGNALR_METHOD = "Broadcast"
SIGNALR_METHOD_ERROR = "Error"
SIGNALR_HUB_FC = "BroadcastHubV2"

FC_GET_OTP = "api/v2/Trading/GetOTP"
FC_GET_ACCESS_TOKEN = "api/v2/Trading/AccessToken"

FC_NEW_ORDER = "api/v2/Trading/NewOrder"
FC_MODIFY_ORDER = "api/v2/Trading/ModifyOrder"
FC_CANCEL_ORDER = "api/v2/Trading/CancelOrder"

FC_DER_NEW_ORDER = "api/v2/Trading/derNewOrder"
FC_DER_MODIFY_ORDER = "api/v2/Trading/derModifyOrder"
FC_DER_CANCEL_ORDER = "api/v2/Trading/derCancelOrder"

FC_GET_ORDER_HISTORY = "api/v2/Trading/orderHistory"
FC_GET_ORDER_BOOK = "api/v2/Trading/orderBook"

FC_GET_DER_POSITION = "api/v2/Trading/derivPosition"
FC_GET_STOCK_POSITION = "api/v2/Trading/stockPosition"
FC_GET_MAX_BUY_QUANTITY = "api/v2/Trading/maxBuyQty"
FC_GET_MAX_SELL_QUANTITY = "api/v2/Trading/maxSellQty"

FC_GET_ACCOUNT_BALANCE = "api/v2/Trading/cashAcctBal"
FC_GET_DER_ACCOUNT_BALANCE = "api/v2/Trading/derivAcctBal"
FC_GET_PP_MMR_ACCOUNT = "api/v2/Trading/ppmmraccount"








import base64
import json
from xml.etree.ElementTree import fromstring
from Crypto.PublicKey import RSA
from Crypto.Signature import pkcs1_15
from Crypto.Hash import SHA256
from xmljson import badgerfish as bf
class core_crypto(object):
	def _init_(self):
		self.ping = 'pong'

	@staticmethod
	def getRSAKey(private_key: str):
		key = base64.b64decode(private_key.encode('utf-8'))
		key_json = json.loads(json.dumps(bf.data(fromstring(key))))
		cc = key_json['RSAKeyValue']
		n = int.from_bytes(base64.b64decode(cc['Modulus']['$']), byteorder='big')
		e = int.from_bytes(base64.b64decode(cc['Exponent']['$']), byteorder='big')
		d = int.from_bytes(base64.b64decode(cc['D']['$']), byteorder='big')
		p = int.from_bytes(base64.b64decode(cc['P']['$']), byteorder='big')
		q = int.from_bytes(base64.b64decode(cc['Q']['$']), byteorder='big')
		#u = int.from_bytes(base64.b64decode(cc['InverseQ']['$']), byteorder='big')
		return RSA.construct((n, e, d, p, q))

	@staticmethod
	def sign(data: str, private_key: str):
		h = SHA256.new(data)
		rsa_key = core_crypto.getRSAKey(private_key)
		return pkcs1_15.new(rsa_key).sign(h).hex()

	@staticmethod
	def sign(data: str, rsa_key: RSA.RsaKey):
		print(data)
		h = SHA256.new(data.encode('utf-8'))
		return pkcs1_15.new(rsa_key).sign(h).hex()




import dataclasses
from .models import fcmodel_requests
from .models import fcmodel_responses
from .models import AccessTokenModel
from .core.core_crypto import core_crypto
from .constant import api
from dataclasses import asdict
import requests
import json
from datetime import date, datetime, timedelta


class FCTradingClient(object):

    def __init__(self, url: str, consumer_id:str, consumer_secret: str, private_key: str, twoFAType: int):
        self.url = url
        self.consumer_id = consumer_id
        self.consumer_secret = consumer_secret
        self.rsa_key = core_crypto.getRSAKey(private_key)
        self._headers = {'Content-Type': 'application/json'}
        self._last_login: datetime = datetime.now()
        self._token_expire_at:  datetime = datetime.now()
        self._twoFAType: int = twoFAType
        self._write_access_token: AccessTokenModel = None
        self._read_access_token: AccessTokenModel = None
        self.get_access_token()

    def _fc_make_request_get(self, _api_url: str, _request_object: object):
        _headers = self._headers
        _headers["Authorization"] = "Bearer " + self.get_access_token()
        tapi_response = requests.get(self.url + _api_url , asdict(_request_object), 
            headers=_headers)

        return tapi_response.json()
        

    def _fc_make_request_post(self, _api_url: str, _request_object: object):
        payload = json.dumps(dataclasses.asdict(_request_object))
        sign_payload = core_crypto.sign(payload, self.rsa_key)
        _headers = self._headers
        _headers["Authorization"] = "Bearer " + self.get_writer_access_token()
        _headers["X-Signature"] = sign_payload
        tapi_response = requests.post(self.url + _api_url , data=payload, 
            headers=_headers)
        return tapi_response.json()

    def get_access_token(self):
        if self._read_access_token == None or self._read_access_token.is_expired():
            req = fcmodel_requests.AccessToken(self.consumer_id, self.consumer_secret, self._twoFAType, '1', False)
            payload = json.dumps(dataclasses.asdict( req))
            res = requests.post(self.url + api.FC_GET_ACCESS_TOKEN , payload, headers=self._headers)
            res_obj = fcmodel_responses.Response(**(res.json()))
            if res_obj.status == 200:
                model = fcmodel_responses.AccessToken(**res_obj.data)
                self._read_access_token = AccessTokenModel(model)
                return self._read_access_token.get_access_token()
            else:
                raise NameError(res_obj.message)
        else:
            return self._read_access_token.get_access_token()

    def get_writer_access_token(self):
        assert self._write_access_token != None, 'Please verify code to get writer token before get it!'
        assert self._write_access_token.is_expired() == False, 'Writer token is expired! Please verify code again to get new one!'
        return self._write_access_token.get_access_token()
    
    def verifyCode(self, code: str):
        req = fcmodel_requests.AccessToken(self.consumer_id, self.consumer_secret, self._twoFAType, code, True)
        payload = json.dumps(dataclasses.asdict( req))
        res = requests.post(self.url + api.FC_GET_ACCESS_TOKEN , payload, headers=self._headers)
        res_obj = fcmodel_responses.Response(**(res.json()))
        if res_obj.status == 200:
            model = fcmodel_responses.AccessToken(**res_obj.data)
            self._write_access_token = AccessTokenModel(model)
            return self._write_access_token.get_access_token()
        else:
            raise NameError(res_obj.message)
    
    def new_order(self, model: fcmodel_requests.NewOrder):
        return self._fc_make_request_post(api.FC_NEW_ORDER, model)

    def modify_order(self, model: fcmodel_requests.ModifyOrder):
        return self._fc_make_request_post(api.FC_MODIFY_ORDER, model)

    def cancle_order(self, model: fcmodel_requests.CancelOrder):
        return self._fc_make_request_post(api.FC_CANCEL_ORDER, model)

    def get_stock_account_balance(self, model: fcmodel_requests.StockAccountBalance):
        return self._fc_make_request_get(api.FC_GET_ACCOUNT_BALANCE, model)

    def get_derivative_account_balance(self, model: fcmodel_requests.DerivativeAccountBalance):
        return self._fc_make_request_get(api.FC_GET_DER_ACCOUNT_BALANCE, model)

    def get_pp_mmr_account(self, model: fcmodel_requests.PPMMRAccount):
        return self._fc_make_request_get(api.FC_GET_PP_MMR_ACCOUNT, model)

    def get_stock_position(self, model: fcmodel_requests.StockPosition):
        return self._fc_make_request_get(api.FC_GET_STOCK_POSITION, model)

    def get_derivative_position(self, model: fcmodel_requests.DerivativePosition):
        return self._fc_make_request_get(api.FC_GET_DER_POSITION, model)

    def get_max_buy_qty(self, model: fcmodel_requests.MaxBuyQty):
        return self._fc_make_request_get(api.FC_GET_MAX_BUY_QUANTITY, model)

    def get_max_sell_qty(self, model: fcmodel_requests.MaxSellQty):
        return self._fc_make_request_get(api.FC_GET_MAX_SELL_QUANTITY, model)

    def get_order_history(self, model: fcmodel_requests.OrderHistory):
        return self._fc_make_request_get(api.FC_GET_ORDER_HISTORY, model)

    def get_otp(self, model: fcmodel_requests.GetOTP):
        payload = json.dumps(dataclasses.asdict(model))
        _headers = self._headers
        tapi_response = requests.post(self.url + api.FC_GET_OTP , data=payload, 
            headers=_headers)
        return tapi_response.json()
    
    def der_new_order(self, model: fcmodel_requests.NewOrder):
        return self._fc_make_request_post(api.FC_DER_NEW_ORDER, model)

    def der_modify_order(self, model: fcmodel_requests.ModifyOrder):
        return self._fc_make_request_post(api.FC_DER_MODIFY_ORDER, model)

    def der_cancle_order(self, model: fcmodel_requests.CancelOrder):
        return self._fc_make_request_post(api.FC_DER_CANCEL_ORDER, model)
    
    def get_order_book(self, model: fcmodel_requests.OrderBook):
        return self._fc_make_request_get(api.FC_GET_ORDER_BOOK, model)


from datetime import datetime
from . import fcmodel_responses
import jwt

class AccessTokenModel(object):

	def __init__(self, access_token: fcmodel_responses.AccessToken):
		jj = jwt.decode(access_token.accessToken, verify=False, options={"verify_signature": False})
		self._token_expire_at = datetime.fromtimestamp(jj['exp'])
		self._access_token = access_token.accessToken

	def get_access_token(self):
		return self._access_token

	def is_expired(self):
		delta = self._token_expire_at - datetime.now()
		if delta.total_seconds() < 3600:
			return True
		return False


import re
from setuptools import setup

# python setup.py sdist bdist_wheel
# twine upload dist/*
def _read_long_description():
    try:
        with open("README.md") as fd:
            return fd.read()
    except Exception:
        return None
    


with open("ssi_fctrading/__init__.py", "r") as fd:
    version = re.search(
        r'^__version__\s*=\s*[\'"]([^\'"]*)[\'"]', fd.read(), re.MULTILINE
    ).group(1)

# I really prefer Markdown to reStructuredText.  PyPi does not.  This allows me
# to have things how I'd like, but not throw complaints when people are trying
# to install the package and they don't have pypandoc or the README in the
# right place.
readme='TBD'


setup(
    name='ssi-fctrading',
    version=version,
    description='FastConnect TradingAPI client by Python',
    long_description=_read_long_description(),
    long_description_content_type='text/markdown',
    author='ducdv',
    author_email='ducdv@ssi.com.vn',
    license='MIT',
    platforms=['POSIX', 'Windows'],
    url='https://github.com/SSI-Securities-Corporation/python-fctrading',
    python_requires=">=3.7",
    classifiers=[
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
    ],
    include_package_data=False,
    install_requires=['pycryptodome',
                        'xmljson',
                        'requests',
                        'signalr-client',
                        'pyjwt'],
)





__version__ = '0.1.1'

# This is a property
# of SSI company
import gevent
from gevent import monkey

monkey.patch_all()
import warnings

warnings.filterwarnings("ignore")
import sys
from .fc_client import FCTradingClient
from .fc_stream import FCTradingStream
